﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;

namespace NetworkConfigurator
{
    /// <summary>
    /// Interaction logic for ConfigurationAreaControl.xaml
    /// </summary>
    public partial class ConfigurationAreaControl : UserControl
    {
        private MainWindow mainWindow;
        private const string LinkControlName = "Link";
        private const float MaxThroughput = 1000; // UoM is Mb/s

        public const int DistanceToBorder = 60;
        public const int IncreaseSize = 500;

        public static bool IsChanged { get; set; } = false;

        public List<Flow> Flows { get; set; }

        public ToolboxControl Source { get; set; }
        public ToolboxControl Destination { get; set; }

        public double ZoomScale
        {
            get { return (double)GetValue(ZoomScaleProperty); }
            set { SetValue(ZoomScaleProperty, value); }
        }

        public static readonly DependencyProperty ZoomScaleProperty =
            DependencyProperty.Register("ZoomScale", typeof(double), typeof(ConfigurationAreaControl), new UIPropertyMetadata());

        public ConfigurationAreaControl()
        {
            InitializeComponent();
            mainWindow = (MainWindow)Application.Current.MainWindow;
            Flows = new List<Flow>();
        }

        public async Task SetFlow()
        {
            int counter = 0, startIndex = 0;
            Dictionary<string, int> elementEncoding = new Dictionary<string, int>(); // <ID:port, counter>
            // Holds the intervals in which the graph is complete.
            Dictionary<int, int> completeGraphIntervals = new Dictionary<int, int>();
            var elements = DrawingArea.Children;
            bool pathExists = false;
            // Construct the graph.
            foreach (var element in elements)
            {
                switch (element.GetType().Name)
                {
                    case "PC":
                        var pc = element as PC;
                        elementEncoding[pc.ID + ":0"] = counter++;
                        //Console.WriteLine(pc.ID + ":0 -> " + (counter - 1));
                        break;
                    case "Switch":
                        var sw = element as Switch;
                        // Remember from which index in the graph this switch begins.
                        startIndex = counter;
                        for (int i = 0; i < sw.NumberOfPorts; i++)
                        {
                            elementEncoding[sw.ID + ":" + i] = counter++;
                            //Console.WriteLine(sw.ID + ":" + i + " -> " + (counter - 1));
                        }
                        completeGraphIntervals[startIndex] = counter;
                        break;
                    case "Router":
                        var r = element as Router;
                        // Remember from which index in the graph this switch begins.
                        startIndex = counter;
                        for (int i = 0; i < r.InterfaceCards.Count; i++)
                        {
                            elementEncoding[r.ID + ":" + i] = counter++;
                            //Console.WriteLine(r.ID + ":" + i + " -> " + (counter - 1));
                        }
                        completeGraphIntervals[startIndex] = counter;
                        break;
                }
            }
            int[,] graph = new int[counter, counter];
            // This matrix contains the links, having as cost the unused bandwidth.
            float[,] costMatrix = new float[counter, counter];
            foreach (var valuePair in completeGraphIntervals)
            {
                for (int i = valuePair.Key; i < valuePair.Value; i++)
                {
                    for (int j = i + 1; j < valuePair.Value; j++)
                    {
                        if (i != j)
                        {
                            graph[i, j] = 1;
                            graph[j, i] = 1;
                            costMatrix[i, j] = MaxThroughput;
                            costMatrix[j, i] = MaxThroughput;
                        }
                    }
                }
            }
            var comm = Flows.Find(c => (c.Source == Source && c.Destination == Destination) ||
                (c.Source == Destination && c.Destination == Source));
            float commThroughput = 0;
            if (comm != null)
                commThroughput = comm.Throughput;
            var links = GetAllLinks();
            foreach (var link in links)
            {
                if (link.IDStart > -1 && link.IDEnd > -1)
                {
                    graph[elementEncoding[link.IDStart + ":" + link.PortIndexStart],
                        elementEncoding[link.IDEnd + ":" + link.PortIndexEnd]] = 1;
                    graph[elementEncoding[link.IDEnd + ":" + link.PortIndexEnd],
                        elementEncoding[link.IDStart + ":" + link.PortIndexStart]] = 1;
                    costMatrix[elementEncoding[link.IDStart + ":" + link.PortIndexStart],
                        elementEncoding[link.IDEnd + ":" + link.PortIndexEnd]] = link.Bandwidth - link.UsedBandwidth;
                    costMatrix[elementEncoding[link.IDEnd + ":" + link.PortIndexEnd],
                        elementEncoding[link.IDStart + ":" + link.PortIndexStart]] = link.Bandwidth - link.UsedBandwidth;
                    //Console.WriteLine(link.IDStart + ":" + link.PortIndexStart);
                    //Console.WriteLine(link.IDEnd + ":" + link.PortIndexEnd);
                }
            }
            // If we are updating this flow then we should add to each link's available bandwidth,
            // the flow's throughtput.
            if(commThroughput > 0)
            {
                links = GetLinksPartOfFlow(comm);
                foreach (var link in links)
                {
                    costMatrix[elementEncoding[link.IDStart + ":" + link.PortIndexStart],
                        elementEncoding[link.IDEnd + ":" + link.PortIndexEnd]] += commThroughput;
                    costMatrix[elementEncoding[link.IDEnd + ":" + link.PortIndexEnd],
                        elementEncoding[link.IDStart + ":" + link.PortIndexStart]] += commThroughput;
                }
            }
            mainWindow.loadingSpinner.Visibility = Visibility.Visible;
            mainWindow.IsEnabled = false;
            mainWindow.MessageTextBlock.Text = "Checking connectivity...";
            // Start a new thread where you do the calculations.
            // By doing so the UI won't block.
            Task t = Task.Run(() =>
            {
                pathExists = IsPathBetween(elementEncoding[Source.ID + ":0"], elementEncoding[Destination.ID + ":0"], graph);
            });

            await t;

            mainWindow.loadingSpinner.Visibility = Visibility.Hidden;
            mainWindow.IsEnabled = true;

            if (pathExists)
            {
                FlowConfigurationWindow w = new FlowConfigurationWindow(Flows, elementEncoding[Source.ID + ":0"],
                    elementEncoding[Destination.ID + ":0"], graph, costMatrix, comm, elementEncoding);
                w.ShowDialog();
                Keyboard.ClearFocus();
                Keyboard.Focus(mainWindow.WidthTextBlock);
                //Console.WriteLine(Keyboard.FocusedElement);
                Source = null;
                Destination = null;
                ResetElementStyle();
            }
            else
            {
                MessageBox.Show("A physical path doesn't exist between the two endpoints.", "Path Finding Error",
                    MessageBoxButton.OK, MessageBoxImage.Exclamation);
            }
        }

        private void DrawingArea_DragOver(object sender, DragEventArgs e)
        {
            e.Effects = DragDropEffects.Copy;
        }

        private void DrawingArea_Drop(object sender, DragEventArgs e)
        {
            // If an element in the panel has already handled the drop,
            // the panel should not also handle it.
            if (e.Handled == false)
            {
                LinkControl ec;
                Canvas canvas = (Canvas)sender;
                var x = e.Data.GetFormats();
                switch (e.Data.GetFormats()[0])
                {
                    case "Object":
                        ToolboxControl toolboxControl = (ToolboxControl)e.Data.GetData("Object");

                        // Get the panel that the element currently belongs to and
                        // add it to the Children of the panel that its been dropped on.
                        Panel _parent = (Panel)VisualTreeHelper.GetParent(toolboxControl);

                        if (_parent != null && e.AllowedEffects.HasFlag(DragDropEffects.Copy))
                        {
                            if (_parent.Name != canvas.Name)
                            {
                                if (toolboxControl.ControlName == LinkControlName)
                                {
                                    LinkControl linkControl = new LinkControl();
                                    Canvas.SetLeft(linkControl.start, e.GetPosition(canvas).X - 100);
                                    Canvas.SetTop(linkControl.start, e.GetPosition(canvas).Y);
                                    Canvas.SetLeft(linkControl.end, e.GetPosition(canvas).X + 100);
                                    Canvas.SetTop(linkControl.end, e.GetPosition(canvas).Y);
                                    canvas.Children.Add(linkControl);
                                    ++LinkControl.NumberOfLinks;
                                }
                                else
                                {
                                    ToolboxControl newToolboxControl = null;
                                    switch (toolboxControl.ControlName)
                                    {
                                        case "PC":
                                            newToolboxControl = new PC(toolboxControl);
                                            break;
                                        case "Switch":
                                            newToolboxControl = new Switch(toolboxControl);
                                            break;
                                        case "Router":
                                            newToolboxControl = new Router(toolboxControl);
                                            break;
                                    }

                                    Canvas.SetTop(newToolboxControl, e.GetPosition((Canvas)sender).Y - newToolboxControl.Height / 2);
                                    Canvas.SetLeft(newToolboxControl, e.GetPosition((Canvas)sender).X - newToolboxControl.Width / 2);
                                    newToolboxControl.AllowDrop = true;
                                    canvas.Children.Add(newToolboxControl);
                                    RearrangeArea(newToolboxControl);
                                }
                            }
                            else
                            {
                                Canvas.SetTop(toolboxControl, e.GetPosition(canvas).Y - toolboxControl.ActualHeight / 2);
                                Canvas.SetLeft(toolboxControl, e.GetPosition(canvas).X - toolboxControl.ActualWidth / 2);
                                RearrangeArea(toolboxControl);
                            }
                        }
                        break;
                    // If dragging is performed on one of the sides then move that side.
                    case "string":
                        ec = (LinkControl)e.Data.GetData("LinkControl");
                        string side = (string)e.Data.GetData("string");
                        if (side == "start")
                        {
                            Canvas.SetLeft(ec.start, e.GetPosition(canvas).X);
                            Canvas.SetTop(ec.start, e.GetPosition(canvas).Y);
                        }
                        else
                        {
                            Canvas.SetLeft(ec.end, e.GetPosition(canvas).X);
                            Canvas.SetTop(ec.end, e.GetPosition(canvas).Y);
                        }
                        break;
                    // Move the entire connector.
                    case MainWindow.LinkControlName:
                        ec = (LinkControl)e.Data.GetData("LinkControl");
                        Canvas.SetLeft(ec.start, e.GetPosition(canvas).X - 50);
                        Canvas.SetTop(ec.start, e.GetPosition(canvas).Y);
                        Canvas.SetLeft(ec.end, e.GetPosition(canvas).X + 50);
                        Canvas.SetTop(ec.end, e.GetPosition(canvas).Y);
                        break;
                }
                IsChanged = true;
            }
        }

        // If the element is close to the border then increase drawing area height/width.
        private void RearrangeArea(ToolboxControl tc)
        {
            var t = (DrawingArea.Height - Canvas.GetTop(tc) - tc.ActualHeight);
            // Measure distance for top and bottom.
            var elements = configAreaCtrl.DrawingArea.Children;
            if (Canvas.GetTop(tc) < DistanceToBorder)
            {
                DrawingArea.Height += IncreaseSize;
                foreach (UIElement element in elements)
                {
                    switch (element.GetType().Name)
                    {
                        case "PC":
                        case "Switch":
                        case "Router":
                            Canvas.SetTop(element, Canvas.GetTop(element) + IncreaseSize);
                            break;
                        case MainWindow.LinkControlName:
                            var ec = element as LinkControl;
                            if (!ec.IsStartConnected)
                                Canvas.SetTop(ec.start, Canvas.GetTop(ec.start) + IncreaseSize);
                            if (!ec.IsEndConnected)
                                Canvas.SetTop(ec.end, Canvas.GetTop(ec.end) + IncreaseSize);
                            break;
                    }
                }
            }
            else if ((DrawingArea.Height - Canvas.GetTop(tc) - tc.ActualHeight) < DistanceToBorder)
            {
                DrawingArea.Height += IncreaseSize;
            }
            // Measure distance for left and right.
            if (Canvas.GetLeft(tc) < DistanceToBorder)
            {
                DrawingArea.Width += IncreaseSize;
                foreach (UIElement element in elements)
                {
                    switch (element.GetType().Name)
                    {
                        case "PC":
                        case "Switch":
                        case "Router":
                            Canvas.SetLeft(element, Canvas.GetLeft(element) + IncreaseSize);
                            break;
                        case MainWindow.LinkControlName:
                            var ec = element as LinkControl;
                            if (!ec.IsStartConnected)
                                Canvas.SetLeft(ec.start, Canvas.GetLeft(ec.start) + IncreaseSize);
                            if (!ec.IsEndConnected)
                                Canvas.SetLeft(ec.end, Canvas.GetLeft(ec.end) + IncreaseSize);
                            break;
                    }
                }
            }
            else if ((DrawingArea.Width - Canvas.GetLeft(tc) - tc.ActualWidth) < 2 * DistanceToBorder)
            {
                DrawingArea.Width += IncreaseSize;
            }
            TrimExtraSpace();
        }

        private void TrimExtraSpace()
        {
            double closestToLeft = DrawingArea.Width, closestToRight = closestToLeft,
                closestToTop = DrawingArea.Height, closestToBottom = closestToTop;
            var elements = DrawingArea.Children;
            foreach (UIElement element in elements)
            {
                if (closestToLeft > Canvas.GetLeft(element))
                    closestToLeft = Canvas.GetLeft(element);
                if (closestToRight > (DrawingArea.Width - Canvas.GetLeft(element)))
                    closestToRight = DrawingArea.Width - Canvas.GetLeft(element);
                if (closestToTop > Canvas.GetTop(element))
                    closestToTop = Canvas.GetTop(element);
                if (closestToBottom > (DrawingArea.Height - Canvas.GetTop(element)))
                    closestToBottom = DrawingArea.Height - Canvas.GetTop(element);
            }
            if (closestToLeft > 600)
            {
                foreach (UIElement element in elements)
                {
                    switch (element.GetType().Name)
                    {
                        case "PC":
                        case "Switch":
                        case "Router":
                            Canvas.SetLeft(element, Canvas.GetLeft(element) - 500);
                            break;
                        case MainWindow.LinkControlName:
                            var ec = element as LinkControl;
                            if (!ec.IsStartConnected)
                                Canvas.SetLeft(ec.start, Canvas.GetLeft(ec.start) - 500);
                            if (!ec.IsEndConnected)
                                Canvas.SetLeft(ec.end, Canvas.GetLeft(ec.end) - 500);
                            break;
                    }
                }
                DrawingArea.Width -= 500;
                ConfigurationAreaControl.IsChanged = true;
            }
            if (closestToRight > 600)
            {
                DrawingArea.Width -= 500;
                ConfigurationAreaControl.IsChanged = true;
            }
            if (closestToTop > 600)
            {
                foreach (UIElement element in elements)
                {
                    switch (element.GetType().Name)
                    {
                        case "PC":
                        case "Switch":
                        case "Router":
                            Canvas.SetTop(element, Canvas.GetTop(element) - 500);
                            break;
                        case MainWindow.LinkControlName:
                            var ec = element as LinkControl;
                            if (!ec.IsStartConnected)
                                Canvas.SetTop(ec.start, Canvas.GetTop(ec.start) - 500);
                            if (!ec.IsEndConnected)
                                Canvas.SetTop(ec.end, Canvas.GetTop(ec.end) - 500);
                            break;
                    }
                }
                DrawingArea.Height -= 500;
                ConfigurationAreaControl.IsChanged = true;
            }
            if (closestToBottom > 600)
            {
                DrawingArea.Height -= 500;
                ConfigurationAreaControl.IsChanged = true;
            }
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            base.OnKeyDown(e);
            if (e.Key == Key.LeftCtrl)
                ChangeElementStyle();
        }

        protected override void OnKeyUp(KeyEventArgs e)
        {
            base.OnKeyUp(e);
            if (e.Key == Key.LeftCtrl)
            {
                Source = null;
                Destination = null;
                ResetElementStyle();
            }
        }

        protected override void OnLostFocus(RoutedEventArgs e)
        {
            base.OnLostFocus(e);
            ResetElementStyle();
        }

        private void ChangeElementStyle()
        {
            var elements = DrawingArea.Children;
            foreach (var element in elements)
            {
                if (element is PC)
                {
                    var tc = element as ToolboxControl;
                    tc.sp.BorderThickness = new Thickness(2);
                }
            }
        }

        public void ResetElementStyle()
        {
            var elements = DrawingArea.Children;
            foreach (var element in elements)
            {
                if (element is PC)
                {
                    var tc = element as ToolboxControl;
                    tc.sp.BorderThickness = new Thickness(1);
                }
            }
        }

        public static bool IsPathBetween(int sourceID, int destinationID, int[,] graph)
        {
            //for (int i = 0; i < graph.GetLength(0); i++)
            //{
            //    for (int j = 0; j < graph.GetLength(1); j++)
            //    {
            //        Console.Write(graph[i, j] + " ");
            //    }
            //    Console.WriteLine();
            //}

            Stack<int> toDoSet = new Stack<int>();
            Stack<int> doneSet = new Stack<int>();
            toDoSet.Push(sourceID);
            while (toDoSet.Count > 0)
            {
                doneSet.Push(toDoSet.Pop());
                for (int i = 0; i < graph.GetLength(1); ++i)
                    if (graph[doneSet.Peek(), i] > 0)
                    {
                        if (i == destinationID)
                            return true;
                        if (!doneSet.Contains(i))
                            toDoSet.Push(i);
                    }
            }
            return false;
        }
        /// <summary>
        /// Return all the links that are connected in
        /// parallel to the link passed in as parameter.
        /// </summary>
        public static List<LinkControl> GetLinksConnectedInParallelToParameter(LinkControl linkParam, UIElementCollection elements)
        {
            var links = new List<LinkControl>();
            // Get all the links that have same IDStart and IDEnd as link.
            foreach (var element in elements)
            {
                if (element is LinkControl && ((((LinkControl)element).IDStart == linkParam.IDStart &&
                    ((LinkControl)element).IDEnd == linkParam.IDEnd) ||
                    (((LinkControl)element).IDStart == linkParam.IDEnd && ((LinkControl)element).IDEnd == linkParam.IDStart)))
                {
                    links.Add(element as LinkControl);
                }
            }
            return links;
        }

        public List<LinkControl> GetAllLinks()
        {
            var elements = DrawingArea.Children;
            var links = new List<LinkControl>();
            foreach (var element in elements)
            {
                if (element is LinkControl)
                    links.Add(element as LinkControl);
            }
            return links;
        }

        public List<LinkControl> GetLinksPartOfFlow(Flow comm)
        {
            var allLinks = GetAllLinks();
            return allLinks.FindAll(l => comm.PathList.Find(t => t.Item1 == l.IDStart && t.Item2 == l.PortIndexStart) != null);
        }

        private void DrawingArea_MouseDown(object sender, MouseButtonEventArgs e)
        {
            mainWindow.HidePath();
        }
    }
}
