﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;

namespace NetworkConfigurator
{
    /// <summary>
    /// Interaction logic for FlowConfigurationWindow.xaml
    /// </summary>
    public partial class FlowConfigurationWindow : Window
    {
        private MainWindow mainWindow;
        private Flow flow;
        private List<Flow> Flows;
        private Brush borderBrush;
        private Dictionary<string, int> elementEncoding;
        // This list contains the decoded ID and port.
        private List<Tuple<int, int>> decodedPath;

        public FlowConfigurationWindow(List<Flow> Flows, int source, int destination, int[,] graph,
            float[,] costMatrix, Flow comm, Dictionary<string, int> elementEncoding)
        {
            InitializeComponent();
            mainWindow = (MainWindow)Application.Current.MainWindow;
            Left = mainWindow.Left + mainWindow.Width / 2 - Width / 2;
            Top = mainWindow.Top + mainWindow.Height / 2 - Height / 2;
            ProtocolsComboBox.ItemsSource = Flow.ApplicationLayerProtocols;
            if (comm != null)
            {
                ThroughputTextBox.Text = comm.Throughput.ToString();
                ProtocolsComboBox.SelectedIndex = Flow.ApplicationLayerProtocols.IndexOf(comm.Protocol);
                IsBidirectionalCheckbox.IsChecked = comm.IsBidirectional;
                this.flow = comm;
                RemoveButton.IsEnabled = true;
            }
            else
            {
                this.flow = new Flow();
                IsBidirectionalCheckbox.IsChecked = true;
                RemoveButton.IsEnabled = false;
            }

            SourceTextBox.Text = mainWindow.configurationArea.Source.ControlName;
            DestinationTextBox.Text = mainWindow.configurationArea.Destination.ControlName;

            this.graph = graph;
            this.costMatrix = costMatrix;
            this.source = source;
            this.destination = destination;
            this.Flows = Flows;
            this.elementEncoding = elementEncoding;
            decodedPath = new List<Tuple<int, int>>();
            borderBrush = ThroughputTextBox.BorderBrush;
        }

        private void ProtocolsComboBox_DropDownClosed(object sender, EventArgs e)
        {
            if (ProtocolsComboBox.SelectedIndex < 0)
            {
                ProtocolsBorder.BorderBrush = Brushes.Red;
                return;
            }

            flow.Protocol = ProtocolsComboBox.SelectedItem.ToString();
            ProtocolsBorder.BorderBrush = null;
        }

        private async void Button_Click(object sender, RoutedEventArgs e)
        {
            if (!IsFormValid())
                return;
            
            mainWindow.loadingSpinner.Visibility = Visibility.Visible;
            mainWindow.IsEnabled = false;
            IsEnabled = false;
            mainWindow.MessageTextBlock.Text = "Calculating route...";

            // Start a new thread where you do the calculations.
            // By doing so the UI won't block.
            Task t = Task.Run(() =>
            {
                // Find the shortest path.
                GetPath();
            });

            await t;

            mainWindow.loadingSpinner.Visibility = Visibility.Hidden;
            mainWindow.IsEnabled = true;
            IsEnabled = true;

            if (encodedPath == null)
            {
                MessageBox.Show("A path has not been found, try lowering the throughput.");
                return;
            }
            // Decode the nodes in the path.
            //Console.WriteLine($"The path from {elementEncoding.First(x => x.Value == source).Key} to " +
                //$"{elementEncoding.First(x => x.Value == destination).Key} is:");
            for (int i = 0; i < encodedPath.Count; i++)
            {
                string key = elementEncoding.First(x => x.Value == encodedPath[i]).Key;
                decodedPath.Add(new Tuple<int, int>(int.Parse(key.Split(':')[0]), int.Parse(key.Split(':')[1])));

                //Console.Write(key + " ");
            }
            var throughputBeforeUpdate = flow.Throughput;
            // This Flow didn't exist before.
            if (flow.Source == null)
            {
                flow.Protocol = ProtocolsComboBox.SelectedItem.ToString();
                flow.Source = mainWindow.configurationArea.Source;
                flow.Destination = mainWindow.configurationArea.Destination;
                flow.Throughput = int.Parse(ThroughputTextBox.Text);
                flow.IsBidirectional = (bool)IsBidirectionalCheckbox.IsChecked;
                flow.PathList = decodedPath;
                Flows.Add(flow);
            }
            else
            {
                flow.Protocol = ProtocolsComboBox.SelectedItem.ToString();
                flow.Throughput = int.Parse(ThroughputTextBox.Text);
                flow.IsBidirectional = (bool)IsBidirectionalCheckbox.IsChecked;
                flow.PathList = decodedPath;
            }
            // Update used bandwidth on each link in the path.
            UpdateUsedBandWidth(throughputBeforeUpdate);
            ConfigurationAreaControl.IsChanged = true;
            DrawPath();
            Close();
        }

        private void DrawPath()
        {
            var links = mainWindow.configurationArea.GetLinksPartOfFlow(flow);
            foreach (var link in links)
            {
                link.startLine.Stroke = Brushes.Green;
                link.endLine.Stroke = Brushes.Green;
            }
        }

        private void UpdateUsedBandWidth(float throughputBeforeUpdate)
        {
            var links = mainWindow.configurationArea.GetAllLinks();
            foreach (var link in links)
            {
                for (int i = 0; i < decodedPath.Count; i++)
                {
                    if (link.IDStart == decodedPath[i].Item1 && link.PortIndexStart == decodedPath[i].Item2)
                    {
                        link.UsedBandwidth += flow.Throughput - throughputBeforeUpdate;
                        break;
                    }
                }
            }
        }

        private void ThroughputTextBox_PreviewTextInput(object sender, TextCompositionEventArgs e)
        {
            e.Handled = !IsValid(ThroughputTextBox.Text + e.Text);
        }

        private bool IsValid(string text)
        {
            if (float.TryParse(text, out float x) && x > 0 && x <= 1000)
                return true;
            else
                return false;
        }

        private void ThroughputTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            if (ThroughputTextBox.Text.Length == 0)
                ThroughputTextBox.BorderBrush = Brushes.Red;
            else
            {
                ThroughputTextBox.BorderBrush = borderBrush;
                Throughput = int.Parse(ThroughputTextBox.Text);
            }
        }

        private bool IsFormValid()
        {
            bool isValid = true;
            if (ProtocolsComboBox.SelectedIndex == -1)
            {
                ProtocolsBorder.BorderBrush = Brushes.Red;
                isValid = false;
            }
            if (ThroughputTextBox.Text.Length == 0)
            {
                ThroughputTextBox.BorderBrush = Brushes.Red;
                isValid = false;
            }

            return isValid;
        }

        private void RemoveButton_Click(object sender, RoutedEventArgs e)
        {
            var linksPartOfFlow = mainWindow.configurationArea.GetLinksPartOfFlow(flow);
            foreach (var link in linksPartOfFlow)
            {
                link.UsedBandwidth -= flow.Throughput;
            }
            Flows.Remove(flow);
            Close();
        }
    }
}
