﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;

namespace NetworkConfigurator
{
    /// <summary>
    /// Interaction logic for LinkControl.xaml
    /// </summary>
    public partial class LinkControl : UserControl, INotifyPropertyChanged
    {
        private static MainWindow mainWindow = (MainWindow)Application.Current.MainWindow;
        private string mouseState;
        private string linkName;
        private Brush borderBrush;
        private string originalName;
        private int offset;
        private int portIndexStart, portIndexEnd;

        public const float BandwidthUpperLimit = 1000f; //1Gb/s
        public const float BandwidthLowerLimit = 0.001f; //

        public event PropertyChangedEventHandler PropertyChanged;

        public bool IsStartConnected { get; set; }
        public bool IsEndConnected { get; set; }
        public int IDStart { get; set; } //  == -1 means it is not connected.
        public int IDEnd { get; set; }
        public int PortIndexStart  //  == -1 means it is not connected.
        {
            get
            {
                return portIndexStart;
            }
            set
            {
                portIndexStart = value;
                OnPropertyChanged();
            }
        }
        public int PortIndexEnd
        {
            get
            {
                return portIndexEnd;
            }
            set
            {
                portIndexEnd = value;
                OnPropertyChanged();
            }
        }

        public static ushort NumberOfLinks = 0;
        // Let C be the point between equipment IDStart and IDEnd
        // This property shows how many units away from C the middle
        // of the link is.
        public int Offset
        {
            get
            {
                return offset;
            }

            set
            {
                offset = value;
                OnPropertyChanged();
            }
        }


        public string LinkName
        {
            get
            {
                return linkName;
            }

            set
            {
                linkName = value;
                OnPropertyChanged();
            }
        }
        public float Bandwidth { get; set; }
        public float UsedBandwidth { get; set; }

        public LinkControl()
        {
            InitializeComponent();
            IDStart = -1;
            IDEnd = -1;
            PortIndexStart = -1;
            PortIndexEnd = -1;
            IsStartConnected = false;
            IsEndConnected = false;
            LinkName = "LINK " + NumberOfLinks;
            borderBrush = LinkNameTextBox.BorderBrush;
            Bandwidth = 100;
        }

        private void OnPropertyChanged([CallerMemberName]string property = "")
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(property));
        }

        private void Ends_MouseMove(object sender, MouseEventArgs e)
        {
            if (Mouse.LeftButton == MouseButtonState.Pressed && mouseState == MainWindow.MouseStateBeforeEnter)
            {
                // Package the data.
                DataObject data = new DataObject();
                data.SetData(MainWindow.LinkControlName, this);
                if (((Rectangle)sender).Name == "start")
                    data.SetData("string", "start");
                else
                    data.SetData("string", "end");

                // Inititate the drag-and-drop operation.
                DragDrop.DoDragDrop(this, data, DragDropEffects.Copy);
            }
        }

        protected override void OnGiveFeedback(GiveFeedbackEventArgs e)
        {
            base.OnGiveFeedback(e);
            // These Effects values are set in the drop target's
            // DragOver event handler.
            if (e.Effects.HasFlag(DragDropEffects.Copy))
                Mouse.SetCursor(Cursors.Cross);
            else
                Mouse.SetCursor(Cursors.No);
            e.Handled = true;
        }

        private void Line_MouseMove(object sender, MouseEventArgs e)
        {
            if (Mouse.LeftButton == MouseButtonState.Pressed && mouseState == MainWindow.MouseStateBeforeEnter &&
                !IsStartConnected && !IsEndConnected)
            {
                // Package the data.
                DataObject data = new DataObject();
                data.SetData(MainWindow.LinkControlName, this);

                // Inititate the drag-and-drop operation.
                DragDrop.DoDragDrop(this, data, DragDropEffects.Copy);
            }
        }

        protected override void OnMouseEnter(MouseEventArgs e)
        {
            base.OnMouseEnter(e);
            if (Mouse.LeftButton == MouseButtonState.Pressed)
                mouseState = "";
            else
                mouseState = MainWindow.MouseStateBeforeEnter;
        }

        // When called it unpluggs the side closest to the mouse pointer.
        // If no side is plugged then nothing happens.
        protected override void OnMouseDoubleClick(MouseButtonEventArgs e)
        {
            base.OnMouseDoubleClick(e);
            if (LinkNameTextBlock.IsMouseOver)
                return;
            var flows = mainWindow.configurationArea.Flows;
            // Check if this link is part of a Flow.
            // true == delete this and the Flow that it is part of.
            // false == stop here.
            // Mouse position has to be saved because after the MessageBox shows up
            // you move the pointer to it and the position is get when you press one of
            // the buttons in it.
            var mousePosition = e.GetPosition((Canvas)Parent);
            if (IsPartOfFlow(flows))
            {
                // Ask user if he wants to proceed.
                var result = MessageBox.Show("This equipment is part of a flow/s.\nIf you delete it the flow/s will also be" +
                    " deleted. Do you want to proceed?", "", MessageBoxButton.YesNo, MessageBoxImage.Question);
                if (result == MessageBoxResult.No)
                    return;
            }

            // Remove the Flow object/s which contain this link.
            RemoveFlows(flows);
            
            var distanceToStart = Math.Sqrt(Math.Pow(Canvas.GetLeft(start) - mousePosition.X, 2) +
            Math.Pow(Canvas.GetTop(start) - mousePosition.Y, 2));
            var distanceToEnd = Math.Sqrt(Math.Pow(Canvas.GetLeft(end) - mousePosition.X, 2) +
                Math.Pow(Canvas.GetTop(end) - mousePosition.Y, 2));
            Random rand = new Random();
            if (distanceToStart < distanceToEnd && IsStartConnected)
            {
                var left = Canvas.GetLeft(start);
                var top = Canvas.GetTop(start);
                // Rearrage other links that might be connected in parallel.
                RearrageLinks();

                var startBoundControl = start.GetBindingExpression(Canvas.LeftProperty).DataItem as ToolboxControl;
                RemoveMeFromEquipment(startBoundControl);

                start.SetBinding(Canvas.LeftProperty, "");
                start.SetBinding(Canvas.TopProperty, "");

                IsStartConnected = false;
                IDStart = -1;
                PortIndexStart = -1;
                // If one side of the link is unplugged then it should be a straight line.
                Offset = 0;

                Canvas.SetLeft(start, left + rand.Next(-50, 50));
                Canvas.SetTop(start, top + rand.Next(25, 50));
            }
            else if (distanceToStart >= distanceToEnd && IsEndConnected)
            {
                var left = Canvas.GetLeft(end);
                var top = Canvas.GetTop(end);
                // Rearrage other links that might be connected in parallel.
                RearrageLinks();

                var endBoundControl = end.GetBindingExpression(Canvas.LeftProperty).DataItem as ToolboxControl;
                RemoveMeFromEquipment(endBoundControl);

                end.SetBinding(Canvas.LeftProperty, "");
                end.SetBinding(Canvas.TopProperty, "");

                IsEndConnected = false;
                IDEnd = -1;
                PortIndexEnd = -1;
                // If one side of the link is unplugged then it should be a straight line.
                Offset = 0;

                Canvas.SetLeft(end, left + rand.Next(-50, 50));
                Canvas.SetTop(end, top + rand.Next(25, 50));
            }
            ConfigurationAreaControl.IsChanged = true;
        }
        /// <summary>
        /// This method rearranges the links that are connected in parallel to this one.
        /// It is called when a link is disconnected or deleted.
        /// </summary>
        private void RearrageLinks()
        {
            var links = ConfigurationAreaControl.GetLinksConnectedInParallelToParameter(this, mainWindow.configurationArea.DrawingArea.Children);
            links.Remove(this);
            if (links.Count == 0)
                return;
            foreach (var link in links)
            {
                if (link.Offset < Offset)
                    ++link.Offset;
            }
            // Find minOffset.
            int minOffset = FindOffset(links, Math.Min);
            // Find maxOffset.
            int maxOffset = FindOffset(links, Math.Max);
            if(minOffset == maxOffset)
            {
                links[0].Offset = 0;
            }                
            while (Math.Abs(maxOffset) - 1 > Math.Abs(minOffset))
            {
                foreach (var link in links)
                {
                    --link.Offset;
                }
                minOffset = FindOffset(links, Math.Max);
                maxOffset = FindOffset(links, Math.Min);
            }
            minOffset = FindOffset(links, Math.Max);
            maxOffset = FindOffset(links, Math.Min);
            while (Math.Abs(minOffset) - 1> Math.Abs(maxOffset))
            {
                foreach (var link in links)
                {
                    ++link.Offset;
                }
                minOffset = FindOffset(links, Math.Max);
                maxOffset = FindOffset(links, Math.Min);
            }
        }

        private int FindOffset(List<LinkControl> links, Func<sbyte, sbyte, sbyte> func)
        {
            int offset = links[0].Offset;
            foreach (var link in links)
            {
                offset = func((sbyte)offset, (sbyte)link.Offset);
            }
            return offset;
        }

        private void RemoveFlows(List<Flow> flows)
        {
            // All Flows that go through this link,
            // Flows that have in their path IDStart and PortIndexStart.
            var myComms = flows.FindAll(c => c.PathList.Find(t => t.Item1 == IDStart && t.Item2 == PortIndexStart) != null);
            
            foreach (var comm in myComms)
            {
                // The other links that form the flow.
                var links = mainWindow.configurationArea.GetLinksPartOfFlow(comm);
                foreach (var link in links)
                {
                    link.UsedBandwidth -= comm.Throughput;
                }
            }
            flows.RemoveAll(c => myComms.Contains(c));
        }

        private bool IsPartOfFlow(List<Flow> flows)
        {
            return flows.Find(c => c.PathList.Find(t => t.Item1 == IDStart) != null &&
                c.PathList.Find(t => t.Item1 == IDEnd) != null) != null ? true : false;
        }

        private void RemoveMeFromEquipment(ToolboxControl boundControl)
        {
            switch (boundControl.GetType().Name)
            {
                case "PC":
                    var pc = boundControl as PC;
                    pc.InterfaceCard.PluggedLink = null;
                    break;
                case "Router":
                    var r = boundControl as Router;
                    foreach (var item in r.InterfaceCards)
                    {
                        if (item.PluggedLink == this)
                        {
                            item.PluggedLink = null;
                            break;
                        }
                    }
                    break;
                case "Switch":
                    var sw = boundControl as Switch;
                    foreach (var link in sw.PortsList)
                    {
                        if (link == this)
                        {
                            if (IDStart == sw.ID)
                                sw.PortsList[link.PortIndexStart] = null;
                            else
                                sw.PortsList[link.PortIndexEnd] = null;
                            --sw.NumberOfOccupiedPorts;
                            break;
                        }
                    }
                    break;
            }
        }

        private void DeleteControl_Click(object sender, RoutedEventArgs e)
        {
            var flows = mainWindow.configurationArea.Flows;
            // true == delete this and the Flow that it is part of.
            // false == stop here.
            if (IsPartOfFlow(flows))
            {
                // Ask user if he wants to proceed.
                var result = MessageBox.Show("This equipment is part of a flow/s.\nIf you delete it the flow/s will also be" +
                    " deleted. Do you want to proceed?", "", MessageBoxButton.YesNo, MessageBoxImage.Question);
                if (result == MessageBoxResult.No)
                    return;
            }

            // Remove the Flow object/s which contain this equipment.
            RemoveFlows(flows);
            var startBoundControl = start.GetBindingExpression(Canvas.LeftProperty);
            if (startBoundControl != null)
                RemoveMeFromEquipment(startBoundControl.DataItem as ToolboxControl);

            var endBoundControl = end.GetBindingExpression(Canvas.LeftProperty);
            if (endBoundControl != null)
                RemoveMeFromEquipment(endBoundControl.DataItem as ToolboxControl);
            RearrageLinks();
            mainWindow.configurationArea.DrawingArea.Children.Remove(this);
        }

        private void EditControl_Click(object sender, RoutedEventArgs e)
        {
            MainWindow window = (MainWindow)Application.Current.MainWindow;
            //Canvas drawingArea = window.configurationArea.DrawingArea;

            //var ec = ((ContextMenu)((MenuItem)sender).CommandParameter).PlacementTarget;

            // Open a window where you set the link name, bandwidth and probably other attributes.
            LinkEditorWindow w = new LinkEditorWindow(this, window);
            w.ShowDialog();
        }

        private void LinkNameTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            TextBox t = sender as TextBox;
            if (t.Text.Length > 0)
                t.BorderBrush = borderBrush;
            else
                t.BorderBrush = Brushes.Red;
        }

        private void TextBlock_MouseDown(object sender, MouseButtonEventArgs e)
        {
            if (e.ClickCount == 2)
            {
                LinkNameTextBox.Visibility = Visibility.Visible;
                LinkNameTextBlock.Visibility = Visibility.Collapsed;
                LinkNameTextBox.Focus();
                LinkNameTextBox.CaretIndex = 9;
                e.Handled = true;
                originalName = LinkName;
            }
        }

        private void LinkNameTextBox_LostFocus(object sender, RoutedEventArgs e)
        {
            LinkNameTextBlock.Visibility = Visibility.Visible;
            LinkNameTextBox.Visibility = Visibility.Collapsed;
            if (LinkNameTextBox.Text == "")
                LinkNameTextBox.Text = originalName;
        }

        public static void SetMyOffset(LinkControl linkParam)
        {
            var links = ConfigurationAreaControl.GetLinksConnectedInParallelToParameter(linkParam,
                mainWindow.configurationArea.DrawingArea.Children);

            // If there are no other links parallel to this one, then it is alright if the link's
            // offset remains 0, a straight line.
            if (links.Count < 2)
                return;
            // Get minOffset and maxOffset from links list.
            int minOffset = links[0].Offset, maxOffset = links[0].Offset;
            foreach (var link in links)
            {
                if (link.Offset < minOffset)
                    minOffset = link.Offset;
                if (link.Offset > maxOffset)
                    maxOffset = link.Offset;
            }
            if (Math.Abs(minOffset) < Math.Abs(maxOffset))
                linkParam.Offset = minOffset - 1;
            else
                linkParam.Offset = maxOffset + 1;
        }
    }
}
