﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Xml;
using System.Xml.Serialization;
using NetworkConfigurator.Converters;
using System.Text.RegularExpressions;
using System.Windows.Media;
using System.Text;

namespace NetworkConfigurator
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>

    public partial class MainWindow : Window
    {
        private const string DrawingAreaTopBottomSide = "TopBottom";
        private const string DrawingAreaLeftRightSide = "LeftRight";

        private string lastSaveLocation = "";

        public const string LinkControlName = "LinkControl";
        public const string ToolboxControlName = "ToolboxControl";
        public const string MouseStateBeforeEnter = "Mouse is in and was NOT previously pressed.";
        

        // The ID of the host/switch as it is in myApp,
        // ex: (h1, 0) meaning h1 has ID 0
        // ex: (sw1, 1) meaning sw1 has ID 1.
        // In my app all IDs are unique between all equipments.
        Dictionary<string, int> myAppIDEncoding;

        public MainWindow()
        {
            InitializeComponent();

            CommandBinding newCommand = new CommandBinding(ApplicationCommands.New);
            CommandBinding saveCommand = new CommandBinding(ApplicationCommands.Save);
            CommandBinding saveAsCommand = new CommandBinding(ApplicationCommands.SaveAs);
            CommandBinding closeCommand = new CommandBinding(ApplicationCommands.Close);
            CommandBinding openCommand = new CommandBinding(ApplicationCommands.Open);

            CommandBindings.Add(newCommand);
            CommandBindings.Add(saveCommand);
            CommandBindings.Add(saveAsCommand);
            CommandBindings.Add(closeCommand);
            CommandBindings.Add(openCommand);

            newCommand.Executed += new ExecutedRoutedEventHandler(NewCommandHandler);
            newCommand.CanExecute += new CanExecuteRoutedEventHandler(NewCommand_CanExecute);

            saveCommand.Executed += new ExecutedRoutedEventHandler(SaveCommandHandler);
            saveCommand.CanExecute += new CanExecuteRoutedEventHandler(SaveCommand_CanExecute);

            saveAsCommand.Executed += new ExecutedRoutedEventHandler(SaveAsCommandHandler);
            saveAsCommand.CanExecute += new CanExecuteRoutedEventHandler(SaveAsCommand_CanExecute);

            closeCommand.Executed += new ExecutedRoutedEventHandler(CloseCommandHandler);
            closeCommand.CanExecute += new CanExecuteRoutedEventHandler(CloseCommand_CanExecute);

            openCommand.Executed += new ExecutedRoutedEventHandler(OpenCommandHandler);
            openCommand.CanExecute += new CanExecuteRoutedEventHandler(OpenCommand_CanExecute);

            IDGenerator.FreeAll();
        }

        private void OpenCommand_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        private void OpenCommandHandler(object sender, ExecutedRoutedEventArgs e)
        {
            switch (e.Parameter as string)
            {
                case "XML":
                case "Mininet":
                    string s = e.Parameter as string;
                    if (ConfigurationAreaControl.IsChanged)
                    {
                        switch (MessageBox.Show("Do you want to save changes?", "Network Configurator", MessageBoxButton.YesNoCancel, MessageBoxImage.Question))
                        {
                            case MessageBoxResult.Yes:
                                SaveConfiguration();
                                if (s == "XML")
                                    OpenXMLConfiguration();
                                else
                                    OpenMininetConfiguration();
                                break;
                            case MessageBoxResult.No:
                                if (s == "XML")
                                    OpenXMLConfiguration();
                                else
                                    OpenMininetConfiguration();
                                break;
                            default:
                                break;
                        }
                    }
                    else
                    {
                        if (s == "XML")
                            OpenXMLConfiguration();
                        else
                            OpenMininetConfiguration();
                    }
                    break;
                case "Help":
                    System.Diagnostics.Process.Start(@".\User Manual.chm");
                    break;
            }            
        }

        private void OpenMininetConfiguration()
        {
            // Configure save file dialog box.
            Microsoft.Win32.OpenFileDialog dlg = new Microsoft.Win32.OpenFileDialog()
            {
                FileName = "Configuration", // Default file name
                DefaultExt = ".txt", // Default file extension
                Filter = "Text documents (.txt)|*.txt" // Filter files by extension
            };

            // Show save file dialog box.
            if (dlg.ShowDialog() == true)
            {
                // Read configuration from dlg.FileName.
                try
                {
                    Reset();
                    ReadMininetConfiguration(dlg.FileName);
                }
                catch (InvalidOperationException e)
                {
                    MessageBox.Show(e.Message, "Network Configurator", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        private void ReadMininetConfiguration(string filePath)
        {
            // Holds a link as it is in the file,
            // ex: "h1-eth1:CS1-eth5"
            List<string> listEncodedLinks;

            int equipmentCounter = 0;
            const string HostRegex = "h[0-9]";
            const string SwitchRegex = "CS[0-9]";
            
            listEncodedLinks = new List<string>();
            myAppIDEncoding = new Dictionary<string, int>();
            
            StreamReader reader = null;
            try
            {
                FileStream fileStream = File.OpenRead(filePath);
                reader = new StreamReader(fileStream);

                string line;
                while ((line = reader.ReadLine()) != null)
                {
                    string[] words = line.Split(' ');
                    // If it is a host.
                    if (Regex.Matches(words[0], HostRegex).Count > 0)
                    {
                        myAppIDEncoding[words[0]] = equipmentCounter++;
                        string[] linkSplit = words[1].Split(':');
                        Array.Sort(linkSplit);
                        listEncodedLinks.Add(linkSplit[0] + ":" + linkSplit[1]);
                    }
                    // If it is a switch.
                    else if (Regex.Matches(words[0], SwitchRegex).Count > 0)
                    {
                        myAppIDEncoding[words[0]] = equipmentCounter++;
                        // i = 0 is "CSxx", i = 1 is "lo:", i = 2 is empty string.
                        for (int i = 3; i < words.Length; i++)
                        {
                            string[] linkSplit = words[i].Split(':');
                            Array.Sort(linkSplit);
                            // If the link is not already counted then we add it.
                            if (!listEncodedLinks.Contains(linkSplit[0] + ":" + linkSplit[1]))
                                listEncodedLinks.Add(linkSplit[0] + ":" + linkSplit[1]);
                        }
                    }
                }

                configurationArea.DrawingArea.Height = (500 + 500 * (equipmentCounter / 4));
                int pcRowCounter = 0, swRowCounter = 0;

                foreach (var eq in myAppIDEncoding)
                {
                    switch (eq.Key[0])
                    {
                        case 'h':
                            var pc = new PC()
                            {
                                ControlName = eq.Key
                            };
                            Canvas.SetLeft(pc, 350);
                            Canvas.SetTop(pc, pcRowCounter++ * 150 + 50);
                            configurationArea.DrawingArea.Children.Add(pc);
                            break;
                        case 'C':
                            var sw = new Switch()
                            {
                                ControlName = eq.Key,
                                MininetID = ParseMininetID(eq.Key)
                            };
                            Canvas.SetLeft(sw, 50);
                            Canvas.SetTop(sw, swRowCounter++ * 150 + 50);
                            configurationArea.DrawingArea.Children.Add(sw);
                            break;
                    }
                }
            }
            catch (IOException)
            {
                MessageBox.Show("An IO exception occured", "", MessageBoxButton.OK, MessageBoxImage.Error);
            }
            catch (Exception)
            {
                MessageBox.Show("An unknown exception occured.", "", MessageBoxButton.OK, MessageBoxImage.Error);
            }
            finally
            {
                reader.Close();
            }

            for (int i = 0; i < listEncodedLinks.Count; ++i)
            {
                string[] sides = listEncodedLinks[i].Split(':');
                // Holds the sides of the link, ex: "h1-eth1".
                string[] equipmetIDAndPort = sides[0].Split('-');
                LinkControl l = new LinkControl()
                {
                    LinkName = "LINK " + i,
                    IDStart = myAppIDEncoding[equipmetIDAndPort[0]],
                    PortIndexStart = int.Parse(equipmetIDAndPort[1].Substring(3)) - 1,
                    Bandwidth = 100,
                    UsedBandwidth = 0,
                    IsStartConnected = true,
                    IsEndConnected = true,
                };

                equipmetIDAndPort = sides[1].Split('-');
                l.IDEnd = myAppIDEncoding[equipmetIDAndPort[0]];
                l.PortIndexEnd = int.Parse(equipmetIDAndPort[1].Substring(3)) - 1;


                // For start side.
                Binding b;

                var tc = FindToolboxControlByID(l.IDStart);

                b = new Binding()
                {
                    Source = tc,
                    Path = new PropertyPath(Canvas.LeftProperty),
                    Mode = BindingMode.TwoWay,
                    Converter = new OffsetConverter(),
                    ConverterParameter = tc.Width / 2
                };
                l.start.SetBinding(Canvas.LeftProperty, b);

                b = new Binding()
                {
                    Source = tc,
                    Path = new PropertyPath(Canvas.TopProperty),
                    Mode = BindingMode.TwoWay,
                    Converter = new OffsetConverter(),
                    ConverterParameter = tc.Height / 2
                };
                l.start.SetBinding(Canvas.TopProperty, b);

                switch (tc.GetType().Name)
                {
                    case "PC":
                        var pc = tc as PC;
                        pc.InterfaceCard.PluggedLink = l;
                        break;
                    case "Switch":
                        var sw = tc as Switch;
                        sw.NumberOfOccupiedPorts++;
                        sw.PortsList[l.PortIndexStart] = l;
                        break;
                    case "Router":
                        var r = tc as Router;
                        r.InterfaceCards[l.PortIndexStart].PluggedLink = l;
                        break;
                }

                // For End side.

                tc = FindToolboxControlByID(l.IDEnd);

                b = new Binding()
                {
                    Source = tc,
                    Path = new PropertyPath(Canvas.LeftProperty),
                    Mode = BindingMode.TwoWay,
                    Converter = new OffsetConverter(),
                    ConverterParameter = tc.Width / 2
                };
                l.end.SetBinding(Canvas.LeftProperty, b);

                b = new Binding()
                {
                    Source = tc,
                    Path = new PropertyPath(Canvas.TopProperty),
                    Mode = BindingMode.TwoWay,
                    Converter = new OffsetConverter(),
                    ConverterParameter = tc.Height / 2
                };
                l.end.SetBinding(Canvas.TopProperty, b);

                switch (tc.GetType().Name)
                {
                    case "PC":
                        var pc = tc as PC;
                        pc.InterfaceCard.PluggedLink = l;
                        break;
                    case "Switch":
                        var sw = tc as Switch;
                        sw.PortsList[l.PortIndexEnd] = l;
                        break;
                    case "Router":
                        var r = tc as Router;
                        r.InterfaceCards[l.PortIndexEnd].PluggedLink = l;
                        break;
                }
                configurationArea.DrawingArea.Children.Add(l);
                LinkControl.SetMyOffset(l);
            }
        }

        private string ParseMininetID(string name)
        {
            StringBuilder number = new StringBuilder(name.Substring(2)),
                id = new StringBuilder();
            for (int i = 0; i < 7; i++)
            {
                if (number.Length > 1)
                {
                    id.Insert(0, ":" + number[number.Length - 2] + number[number.Length - 1]);
                    number.Remove(number.Length - 2, 2);
                }
                else if(number.Length > 0)
                {
                    id.Insert(0, ":0" + number[number.Length - 1]);
                    number.Remove(number.Length - 1, 1);
                }
                else
                {
                    id.Insert(0, ":00");
                }                
            }
            if (number.Length > 1)
            {
                id.Insert(0, number[number.Length - 2] + number[number.Length - 1]);
            }
            else if (number.Length > 0)
            {
                id.Insert(0, "0" + number.Remove(number.Length - 1, 1));
            }
            else
            {
                id.Insert(0, "00");
            }
            return id.ToString();
        }

        private void OpenXMLConfiguration()
        {
            // Configure save file dialog box.
            Microsoft.Win32.OpenFileDialog dlg = new Microsoft.Win32.OpenFileDialog()
            {
                FileName = "Configuration", // Default file name
                DefaultExt = ".xml", // Default file extension
                Filter = "Text documents (.xml)|*.xml" // Filter files by extension
            };

            // Show save file dialog box.
            if (dlg.ShowDialog() == true)
            {
                // Read configuration from dlg.FileName.
                try
                {
                    var configuration = ReadXMLConfiguration(dlg.FileName);
                    Reset();

                    configAreaWidth.Text = configuration.ConfigurationAreaWidth.ToString();
                    configAreaHeight.Text = configuration.ConfigurationAreaHeight.ToString();
                    configurationArea.DrawingArea.Height = configuration.ConfigurationAreaHeight;
                    configurationArea.DrawingArea.Width = configuration.ConfigurationAreaWidth;
                    uiScaleSlider.Value = configuration.ZoomScale;

                    // For all PCs.
                    foreach (var tcObj in configuration.ToolboxControls.PCs.PCsList)
                    {
                        var pc = new PC
                        {
                            ID = tcObj.ID,
                            ControlName = tcObj.ControlName,
                            ImageSource = tcObj.ImageSource,
                            InterfaceCard = new NetworkInterfaceCard
                            {
                                Index = tcObj.NIC.Index,
                                IPAddress = tcObj.NIC.IPAddress,
                                MACAddress = tcObj.NIC.MACAddress
                            }
                        };

                        Canvas.SetLeft(pc, tcObj.CanvasLeft);
                        Canvas.SetTop(pc, tcObj.CanvasTop);

                        configurationArea.DrawingArea.Children.Add(pc);
                    }

                    // For all Switches.
                    foreach (var tcObj in configuration.ToolboxControls.Switches.SwitchesList)
                    {
                        var sw = new Switch
                        {
                            ID = tcObj.ID,
                            ControlName = tcObj.ControlName,
                            ImageSource = tcObj.ImageSource,
                            MininetID = tcObj.MininetID,
                            NumberOfOccupiedPorts = tcObj.NumberOfOccupiedPorts,
                            NumberOfPorts = tcObj.NumberOfPorts,
                            PortsList = new List<LinkControl>()
                        };
                        for (int i = 0; i < sw.NumberOfPorts; i++)
                        {
                            sw.PortsList.Add(null);
                        }
                        Canvas.SetLeft(sw, tcObj.CanvasLeft);
                        Canvas.SetTop(sw, tcObj.CanvasTop);

                        configurationArea.DrawingArea.Children.Add(sw);
                    }

                    // For all Routers.
                    foreach (var tcObj in configuration.ToolboxControls.Routers.RoutersList)
                    {
                        var router = new Router
                        {
                            ID = tcObj.ID,
                            ControlName = tcObj.ControlName,
                            ImageSource = tcObj.ImageSource,
                            InterfaceCards = new System.Collections.ObjectModel.ObservableCollection<NetworkInterfaceCard>()
                        };

                        foreach (var nic in tcObj.NetworkInterfaceCardsObject.NICs)
                        {
                            router.InterfaceCards.Add(new NetworkInterfaceCard
                            {
                                Index = nic.Index,
                                IPAddress = nic.IPAddress,
                                MACAddress = nic.MACAddress
                            });
                        }

                        Canvas.SetLeft(router, tcObj.CanvasLeft);
                        Canvas.SetTop(router, tcObj.CanvasTop);

                        configurationArea.DrawingArea.Children.Add(router);
                    }

                    // For all links.
                    var elements = configurationArea.DrawingArea.Children;

                    foreach (var ecObj in configuration.LinkControls.LinkControlsList)
                    {
                        var ec = new LinkControl
                        {
                            LinkName = ecObj.LinkName,
                            IDStart = ecObj.IDStart,
                            IDEnd = ecObj.IDEnd,
                            PortIndexStart = ecObj.PortIndexStart,
                            PortIndexEnd = ecObj.PortIndexEnd,
                            Bandwidth = ecObj.Bandwidth,
                            UsedBandwidth = ecObj.UsedBandwidth,
                            IsStartConnected = ecObj.IsStartConnected,
                            IsEndConnected = ecObj.IsEndConnected,
                            Offset = ecObj.Offset
                        };

                        // For start side.
                        if (ec.IsStartConnected)
                        {
                            Binding b;

                            var tc = FindToolboxControlByID(ec.IDStart);

                            b = new Binding()
                            {
                                Source = tc,
                                Path = new PropertyPath(Canvas.LeftProperty),
                                Mode = BindingMode.TwoWay,
                                Converter = new OffsetConverter(),
                                ConverterParameter = tc.Width / 2
                            };
                            ec.start.SetBinding(Canvas.LeftProperty, b);

                            b = new Binding()
                            {
                                Source = tc,
                                Path = new PropertyPath(Canvas.TopProperty),
                                Mode = BindingMode.TwoWay,
                                Converter = new OffsetConverter(),
                                ConverterParameter = tc.Height / 2
                            };
                            ec.start.SetBinding(Canvas.TopProperty, b);

                            switch (tc.GetType().Name)
                            {
                                case "PC":
                                    var pc = tc as PC;
                                    pc.InterfaceCard.PluggedLink = ec;
                                    break;
                                case "Switch":
                                    var sw = tc as Switch;
                                    sw.PortsList[ec.PortIndexStart] = ec;
                                    break;
                                case "Router":
                                    var r = tc as Router;
                                    r.InterfaceCards[ec.PortIndexStart].PluggedLink = ec;
                                    break;
                            }
                        }
                        else
                        {
                            Canvas.SetLeft(ec.start, ecObj.StartCanvasLeft);
                            Canvas.SetTop(ec.start, ecObj.StartCanvasTop);
                        }

                        // For End side.
                        if (ec.IsEndConnected)
                        {
                            Binding b;

                            var tc = FindToolboxControlByID(ec.IDEnd);

                            b = new Binding()
                            {
                                Source = tc,
                                Path = new PropertyPath(Canvas.LeftProperty),
                                Mode = BindingMode.TwoWay,
                                Converter = new OffsetConverter(),
                                ConverterParameter = tc.Width / 2
                            };
                            ec.end.SetBinding(Canvas.LeftProperty, b);

                            b = new Binding()
                            {
                                Source = tc,
                                Path = new PropertyPath(Canvas.TopProperty),
                                Mode = BindingMode.TwoWay,
                                Converter = new OffsetConverter(),
                                ConverterParameter = tc.Height / 2
                            };
                            ec.end.SetBinding(Canvas.TopProperty, b);

                            switch (tc.GetType().Name)
                            {
                                case "PC":
                                    var pc = tc as PC;
                                    pc.InterfaceCard.PluggedLink = ec;
                                    break;
                                case "Switch":
                                    var sw = tc as Switch;
                                    sw.PortsList[ec.PortIndexEnd] = ec;
                                    break;
                                case "Router":
                                    var r = tc as Router;
                                    r.InterfaceCards[ec.PortIndexEnd].PluggedLink = ec;
                                    break;
                            }
                        }
                        else
                        {
                            Canvas.SetLeft(ec.end, ecObj.EndCanvasLeft);
                            Canvas.SetTop(ec.end, ecObj.EndCanvasTop);
                        }

                        configurationArea.DrawingArea.Children.Add(ec);
                    }

                    // For all Flows.
                    configurationArea.Flows = new List<Flow>();
                    foreach (var comm in configuration.FlowsObject.FlowsList)
                    {
                        var path = new List<Tuple<int, int>>();
                        foreach (var vertex in comm.Path.Value)
                        {
                            path.Add(new Tuple<int, int>(vertex.EquipmentID, vertex.PortID));
                        }
                        configurationArea.Flows.Add(new Flow
                        {
                            Protocol = comm.Protocol,
                            Source = FindToolboxControlByID(comm.SourceID),
                            Destination = FindToolboxControlByID(comm.DestinationID),
                            Throughput = comm.Throughput,
                            IsBidirectional = bool.Parse(comm.IsBidirectional),
                            PathList = path
                        });
                    }
                }
                catch (InvalidOperationException e)
                {
                    MessageBox.Show(e.Message, "Network Configurator", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        public ToolboxControl FindToolboxControlByID(int ID)
        {
            var elements = configurationArea.DrawingArea.Children;
            foreach (var element in elements)
            {
                if (element is ToolboxControl)
                {
                    var tc = element as ToolboxControl;
                    if (tc.ID == ID)
                        return tc;
                }
            }
            return null;
        }

        private Configuration ReadXMLConfiguration(string fileName)
        {
            XmlSerializer serializer = new XmlSerializer(typeof(Configuration));
            using (TextReader reader = new StringReader(File.ReadAllText(fileName)))
            {
                try
                {
                    return (Configuration)serializer.Deserialize(reader);
                }
                catch (InvalidOperationException e)
                {
                    throw new InvalidOperationException(e.Message);
                }
                catch (Exception e)
                {
                    throw new Exception(e.Message);
                }
            }
        }

        private void CloseCommand_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        private void CloseCommandHandler(object sender, ExecutedRoutedEventArgs e)
        {
            Close();
        }

        private void SaveAsCommandHandler(object sender, ExecutedRoutedEventArgs e)
        {
            SaveConfiguration();
        }

        private void SaveAsCommand_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            if (configurationArea.DrawingArea.Children.Count > 0)
                e.CanExecute = true;
            else
                e.CanExecute = false;
        }

        private void SaveCommand_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = ConfigurationAreaControl.IsChanged ? true : false;
        }

        private void SaveCommandHandler(object sender, ExecutedRoutedEventArgs e)
        {
            if (lastSaveLocation == "")
                SaveConfiguration();
            else
                WriteConfiguration(lastSaveLocation);
            ConfigurationAreaControl.IsChanged = false;
        }

        private void NewCommand_CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            if (configurationArea.DrawingArea.Children.Count > 0)
                e.CanExecute = true;
            else
                e.CanExecute = false;
        }

        private void NewCommandHandler(object sender, ExecutedRoutedEventArgs e)
        {
            // Check if the drawing area is changed.
            if (ConfigurationAreaControl.IsChanged)
            {
                switch (MessageBox.Show("Do you want to save changes?", "Network Configurator", MessageBoxButton.YesNoCancel, MessageBoxImage.Question))
                {
                    case MessageBoxResult.Yes:
                        // Open a dialog box.
                        SaveConfiguration();
                        // Reset configuration area.
                        Reset();
                        break;
                    case MessageBoxResult.No:
                        // Reset the configuration area.
                        Reset();
                        break;
                    default:
                        break;
                }
            }
            else
                Reset();
        }

        private void Reset()
        {
            configurationArea.DrawingArea.Children.Clear();
            configurationArea.DrawingArea.Height = 500;
            configurationArea.DrawingArea.Width = 500;
            configAreaHeight.Text = "500";
            configAreaWidth.Text = "500";
            configurationArea.DrawingArea.Height = 500;
            configurationArea.Flows.Clear();
            uiScaleSlider.Value = 1;
            ConfigurationAreaControl.IsChanged = false;
            lastSaveLocation = "";
            IDGenerator.FreeAll();
            LinkControl.NumberOfLinks = 0;
        }

        private void SaveConfiguration()
        {
            // Configure save file dialog box.
            Microsoft.Win32.SaveFileDialog dlg = new Microsoft.Win32.SaveFileDialog()
            {
                FileName = "Configuration", // Default file name
                DefaultExt = ".xml", // Default file extension
                Filter = "Text documents (.xml)|*.xml" // Filter files by extension
            };

            // Show save file dialog box.
            if (dlg.ShowDialog() == true)
            {
                // Write configuration to dlg.FileName.
                WriteConfiguration(dlg.FileName);
                lastSaveLocation = dlg.FileName;
                ConfigurationAreaControl.IsChanged = false;
            }
        }

        private void WriteConfiguration(string fileName)
        {
            XmlWriterSettings settings = new XmlWriterSettings()
            {
                Indent = true,
                IndentChars = "\t"
            };
            XmlWriter writer = XmlWriter.Create(fileName, settings);

            writer.WriteStartElement("Configuration");
            writer.WriteElementString("ConfigurationAreaWidth", configAreaWidth.Text);
            writer.WriteElementString("ConfigurationAreaHeight", configAreaHeight.Text);
            writer.WriteElementString("ZoomScale", uiScaleSlider.Value.ToString("F"));

            var elements = configurationArea.DrawingArea.Children;

            writer.WriteStartElement("ToolboxControls");

            writer.WriteStartElement("PCs");
            // For all PCs.
            foreach (var element in elements)
            {
                if (element.GetType().Name == typeof(PC).Name)
                {
                    var pc = element as PC;
                    writer.WriteStartElement("PC");
                    writer.WriteElementString("ID", pc.ID.ToString());
                    writer.WriteElementString("ControlName", pc.ControlName);
                    writer.WriteElementString("ImageSource", pc.ImageSource);
                    writer.WriteElementString("CanvasLeft", Canvas.GetLeft(pc).ToString("F"));
                    writer.WriteElementString("CanvasTop", Canvas.GetTop(pc).ToString("F"));
                    // Write the PC's NIC.
                    writer.WriteStartElement("NetworkInterfaceCard");
                    writer.WriteElementString("Index", pc.InterfaceCard.Index.ToString());
                    writer.WriteElementString("IPAddress", pc.InterfaceCard.IPAddress);
                    writer.WriteElementString("MACAddress", pc.InterfaceCard.MACAddress);
                    // The plugged links is read from the links list.
                    // End of PC NIC.
                    writer.WriteEndElement();
                    writer.WriteEndElement();
                }
            }
            // End PCs.
            writer.WriteEndElement();

            writer.WriteStartElement("Switches");
            // For all Switches.
            foreach (var element in elements)
            {
                if (element.GetType().Name == typeof(Switch).Name)
                {
                    var sw = element as Switch;
                    writer.WriteStartElement("Switch");
                    writer.WriteElementString("ID", sw.ID.ToString());
                    writer.WriteElementString("ControlName", sw.ControlName);
                    writer.WriteElementString("ImageSource", sw.ImageSource);
                    writer.WriteElementString("MininetID", sw.MininetID);
                    writer.WriteElementString("CanvasLeft", Canvas.GetLeft(sw).ToString("F"));
                    writer.WriteElementString("CanvasTop", Canvas.GetTop(sw).ToString("F"));
                    writer.WriteElementString("NumberOfPorts", sw.NumberOfPorts.ToString());
                    writer.WriteElementString("NumberOfOccupiedPorts", sw.NumberOfOccupiedPorts.ToString());
                    // The links are read from the links list.
                    writer.WriteEndElement();
                }
            }
            // End Switches.
            writer.WriteEndElement();

            writer.WriteStartElement("Routers");
            // For all Routers.
            foreach (var element in elements)
            {
                if (element.GetType().Name == typeof(Router).Name)
                {
                    var router = element as Router;
                    writer.WriteStartElement("Router");
                    writer.WriteElementString("ID", router.ID.ToString());
                    writer.WriteElementString("ControlName", router.ControlName);
                    writer.WriteElementString("ImageSource", router.ImageSource);
                    writer.WriteElementString("CanvasLeft", Canvas.GetLeft(router).ToString("F"));
                    writer.WriteElementString("CanvasTop", Canvas.GetTop(router).ToString("F"));
                    // For all NICs.
                    writer.WriteStartElement("NetworkInterfaceCards");
                    foreach (var nic in router.InterfaceCards)
                    {
                        writer.WriteStartElement("NetworkInterfaceCard");
                        writer.WriteElementString("Index", nic.Index.ToString());
                        writer.WriteElementString("IPAddress", nic.IPAddress);
                        writer.WriteElementString("MACAddress", nic.MACAddress);
                        // The plugged link is assigned from the links list.
                        writer.WriteEndElement();
                    }
                    // End NICs.
                    writer.WriteEndElement();
                    writer.WriteEndElement();
                }
            }
            // End Routers.
            writer.WriteEndElement();

            // End ToolboxControls.
            writer.WriteEndElement();

            writer.WriteStartElement("LinkControls");
            // For all links.
            foreach (var element in elements)
            {
                if (element.GetType().Name == LinkControlName)
                {
                    var ec = element as LinkControl;
                    writer.WriteStartElement(LinkControlName);
                    writer.WriteElementString("LinkName", ec.LinkName);
                    writer.WriteElementString("IDStart", ec.IDStart.ToString());
                    writer.WriteElementString("IDEnd", ec.IDEnd.ToString());
                    writer.WriteElementString("PortIndexStart", ec.PortIndexStart.ToString());
                    writer.WriteElementString("PortIndexEnd", ec.PortIndexEnd.ToString());
                    writer.WriteElementString("Bandwidth", ec.Bandwidth.ToString("F"));
                    writer.WriteElementString("UsedBandwidth", ec.UsedBandwidth.ToString("F"));
                    writer.WriteElementString("StartCanvasLeft", Canvas.GetLeft(ec.start).ToString("F"));
                    writer.WriteElementString("StartCanvasTop", Canvas.GetTop(ec.start).ToString("F"));
                    writer.WriteElementString("EndCanvasLeft", Canvas.GetLeft(ec.end).ToString("F"));
                    writer.WriteElementString("EndCanvasTop", Canvas.GetTop(ec.end).ToString("F"));
                    writer.WriteElementString("IsStartConnected", ec.IsStartConnected.ToString().ToLower());
                    writer.WriteElementString("IsEndConnected", ec.IsEndConnected.ToString().ToLower());
                    writer.WriteElementString("Offset", ec.Offset.ToString());
                    writer.WriteEndElement();
                }
            }
            // End LinkControls.
            writer.WriteEndElement();

            // For all Flows
            writer.WriteStartElement("Flows");
            foreach (var comm in configurationArea.Flows)
            {
                writer.WriteStartElement("Flow");
                writer.WriteElementString("Protocol", comm.Protocol);
                writer.WriteElementString("SourceID", comm.Source.ID.ToString());
                writer.WriteElementString("DestinationID", comm.Destination.ID.ToString());
                writer.WriteElementString("Throughput", comm.Throughput.ToString());
                writer.WriteElementString("IsBidirectional", comm.IsBidirectional.ToString());
                // Path
                writer.WriteStartElement("Path");
                foreach (var vertex in comm.PathList)
                {
                    writer.WriteStartElement("Vertex");
                    writer.WriteElementString("EquipmentID", vertex.Item1.ToString());
                    writer.WriteElementString("PortID", vertex.Item2.ToString());
                    writer.WriteEndElement();
                }
                // End Path.
                writer.WriteEndElement();
                // End Flow.
                writer.WriteEndElement();
            }
            // End Flows.
            writer.WriteEndElement();

            // End Configuration.
            writer.WriteEndElement();

            writer.Close();
        }

        public void UiScaleSlider_MouseDoubleClick(object sender, MouseButtonEventArgs e)
        {
            ((Slider)sender).Value = 1;
            ConfigurationAreaControl.IsChanged = true;
        }

        private void SliderMinus_Click(object sender, RoutedEventArgs e)
        {
            uiScaleSlider.Value -= 0.2;
            ConfigurationAreaControl.IsChanged = true;
        }

        private void SliderPlus_Click_1(object sender, RoutedEventArgs e)
        {
            uiScaleSlider.Value += 0.2;
            ConfigurationAreaControl.IsChanged = true;
        }

        private void WidthTextBox_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.Key != Key.Return)
                return;

            var t = sender as TextBox;
            ValidateConfigurationAreaWidth(t);
        }

        private void ConfigAreaWidth_LostFocus(object sender, RoutedEventArgs e)
        {
            ValidateConfigurationAreaWidth(sender as TextBox);
        }

        private void ValidateConfigurationAreaWidth(TextBox t)
        {
            if (Int16.TryParse(t.Text, out short newWidth))
            {
                if (newWidth < configurationArea.DrawingArea.Width &&
                    IsValidSize(DrawingAreaLeftRightSide, ((int)configurationArea.DrawingArea.Width - newWidth) / 2))
                {
                    configurationArea.DrawingArea.Width = newWidth;
                    ConfigurationAreaControl.IsChanged = true;
                }
                else
                    t.Text = configurationArea.DrawingArea.Width.ToString();
                if (newWidth > configurationArea.DrawingArea.Width)
                {
                    foreach (UIElement element in configurationArea.DrawingArea.Children)
                    {
                        switch (element.GetType().Name)
                        {
                            case "PC":
                            case "Switch":
                            case "Router":
                                Canvas.SetLeft(element, Canvas.GetLeft(element) + (newWidth - configurationArea.DrawingArea.Width) / 2);
                                break;
                            case LinkControlName:
                                var ec = element as LinkControl;
                                if (!ec.IsStartConnected)
                                    Canvas.SetLeft(ec.start, Canvas.GetLeft(ec.start) + (newWidth - configurationArea.DrawingArea.Width) / 2);
                                if (!ec.IsEndConnected)
                                    Canvas.SetLeft(ec.end, Canvas.GetLeft(ec.end) + (newWidth - configurationArea.DrawingArea.Width) / 2);
                                break;
                        }
                    }
                    configurationArea.DrawingArea.Width = newWidth;
                    ConfigurationAreaControl.IsChanged = true;
                }
            }
            else
                t.Text = configurationArea.DrawingArea.Width.ToString();
        }

        private void HeightTextBox_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.Key != Key.Return)
                return;
            var t = sender as TextBox;
            ValidateConfigurationAreaHeight(t);
        }

        private void ConfigAreaHeight_LostFocus(object sender, RoutedEventArgs e)
        {
            ValidateConfigurationAreaHeight(sender as TextBox);
        }

        private void ValidateConfigurationAreaHeight(TextBox t)
        {
            if (Int16.TryParse(t.Text, out short newHeight))
            {
                if (newHeight < configurationArea.DrawingArea.Height &&
                    IsValidSize(DrawingAreaTopBottomSide, ((int)configurationArea.DrawingArea.Height - newHeight) / 2))
                {
                    configurationArea.DrawingArea.Height = newHeight;
                    ConfigurationAreaControl.IsChanged = true;
                }
                else
                    t.Text = configurationArea.DrawingArea.Height.ToString();
                if (newHeight > configurationArea.DrawingArea.Height)
                {
                    foreach (UIElement element in configurationArea.DrawingArea.Children)
                    {
                        switch (element.GetType().Name)
                        {
                            case "PC":
                            case "Switch":
                            case "Router":
                                Canvas.SetTop(element, Canvas.GetTop(element) + (newHeight - configurationArea.DrawingArea.Height) / 2);
                                break;
                            case LinkControlName:
                                var ec = element as LinkControl;
                                if (!ec.IsStartConnected)
                                    Canvas.SetTop(ec.start, Canvas.GetTop(ec.start) + (newHeight - configurationArea.DrawingArea.Height) / 2);
                                if (!ec.IsEndConnected)
                                    Canvas.SetTop(ec.end, Canvas.GetTop(ec.end) + (newHeight - configurationArea.DrawingArea.Height) / 2);
                                break;
                        }
                    }
                    configurationArea.DrawingArea.Height = newHeight;
                    ConfigurationAreaControl.IsChanged = true;
                }
            }
            else
                t.Text = configurationArea.DrawingArea.Height.ToString();
        }

        private bool IsValidSize(string side, int size)
        {
            var elements = configurationArea.DrawingArea.Children;
            switch (side)
            {
                case DrawingAreaLeftRightSide:
                    // Check distance to left and right borders.
                    foreach (UIElement element in elements)
                    {
                        switch (element.GetType().Name)
                        {
                            case "PC":
                            case "Switch":
                            case "Router":
                                var tc = element as ToolboxControl;
                                if (Canvas.GetLeft(tc) < size ||
                                    (Canvas.GetLeft(tc) + tc.ActualWidth > configurationArea.DrawingArea.Width - size))
                                {
                                    return false;
                                }
                                break;
                            case LinkControlName:
                                var ec = element as LinkControl;
                                if (Canvas.GetLeft(ec.start) < size || Canvas.GetLeft(ec.end) < size ||
                                    (Canvas.GetLeft(ec.start) + ec.start.ActualWidth > configurationArea.DrawingArea.Width - size) ||
                                    (Canvas.GetLeft(ec.end) + ec.end.ActualWidth > configurationArea.DrawingArea.Width - size))
                                {
                                    return false;
                                }
                                break;
                        }
                    }
                    // Move each element to the top with the value of "size".
                    foreach (UIElement element in elements)
                    {
                        switch (element.GetType().Name)
                        {
                            case "PC":
                            case "Switch":
                            case "Router":
                                Canvas.SetLeft(element, Canvas.GetLeft(element) - size);
                                break;
                            case LinkControlName:
                                var ec = element as LinkControl;
                                if (!ec.IsStartConnected)
                                    Canvas.SetLeft(ec.start, Canvas.GetLeft(ec.start) - size);
                                if (!ec.IsEndConnected)
                                    Canvas.SetLeft(ec.end, Canvas.GetLeft(ec.end) - size);
                                break;
                        }
                    }
                    break;
                case DrawingAreaTopBottomSide:
                    // Check distance to top and bottom borders.
                    foreach (UIElement element in elements)
                    {
                        switch (element.GetType().Name)
                        {
                            case "PC":
                            case "Switch":
                            case "Router":
                                var tc = element as ToolboxControl;
                                if (Canvas.GetTop(tc) < size ||
                                    (Canvas.GetTop(tc) + tc.ActualHeight > configurationArea.DrawingArea.Height - size))
                                {
                                    return false;
                                }
                                break;
                            case LinkControlName:
                                var ec = element as LinkControl;
                                if (Canvas.GetTop(ec.start) < size || Canvas.GetTop(ec.end) < size ||
                                    (Canvas.GetTop(ec.start) + ec.start.ActualHeight > configurationArea.DrawingArea.Height - size) ||
                                    (Canvas.GetTop(ec.end) + ec.end.ActualHeight > configurationArea.DrawingArea.Height - size))
                                {
                                    return false;
                                }
                                break;
                        }
                    }
                    // Move each element to the top with the value of "size".
                    foreach (UIElement element in elements)
                    {
                        switch (element.GetType().Name)
                        {
                            case "PC":
                            case "Switch":
                            case "Router":

                                Canvas.SetTop(element, Canvas.GetTop(element) - size);
                                break;
                            case LinkControlName:
                                var ec = element as LinkControl;
                                if (!ec.IsStartConnected)
                                    Canvas.SetTop(ec.start, Canvas.GetTop(ec.start) - size);
                                if (!ec.IsEndConnected)
                                    Canvas.SetTop(ec.end, Canvas.GetTop(ec.end) - size);
                                break;
                        }
                    }
                    break;
            }
            return true;
        }

        private void ClosingHandler(object sender, System.ComponentModel.CancelEventArgs e)
        {
            if (configurationArea.DrawingArea.Children.Count == 0)
                return;

            // Check if the drawing area is changed.
            if (ConfigurationAreaControl.IsChanged)
            {
                switch (MessageBox.Show("Do you want to save changes before exiting?", "Network Configurator", MessageBoxButton.YesNoCancel, MessageBoxImage.Question))
                {
                    case MessageBoxResult.Yes:
                        // Open a dialog box.
                        SaveConfiguration();
                        break;
                    case MessageBoxResult.No:
                        // Close the application.
                        break;
                    default:
                        e.Cancel = true;
                        break;
                }
            }
        }

        public void HidePath()
        {
            var links = configurationArea.GetAllLinks();
            foreach (var link in links)
            {
                link.startLine.Stroke = Brushes.Goldenrod;
                link.endLine.Stroke = Brushes.Goldenrod;
            }
        }
    }
}
