﻿using NetworkConfigurator.Converters;
using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Threading;
using System.Collections.ObjectModel;
using System.Collections.Generic;
using System.Threading.Tasks;

namespace NetworkConfigurator
{
    /// <summary>
    /// Interaction logic for ToolboxControl.xaml
    /// </summary>
    /// 

    public partial class ToolboxControl : UserControl
    {
        private static MainWindow mainWindow = (MainWindow)Application.Current.MainWindow;
        private string mouseState;
        private Brush borderBrush;
        private string originalName;
        public int ID { get; set; }

        public string ImageSource
        {
            get { return (string)GetValue(ImageSourceProperty); }
            set { SetValue(ImageSourceProperty, value); }
        }

        public static readonly DependencyProperty ImageSourceProperty =
            DependencyProperty.Register("ImageSource", typeof(string), typeof(ToolboxControl), new UIPropertyMetadata());

        public string ControlName
        {
            get { return (string)GetValue(ControlNameProperty); }
            set { SetValue(ControlNameProperty, value); }
        }

        public static readonly DependencyProperty ControlNameProperty =
            DependencyProperty.Register("ControlName", typeof(string), typeof(ToolboxControl), new UIPropertyMetadata());

        public ToolboxControl()
        {
            InitializeComponent();
            ID = IDGenerator.NextID();
            borderBrush = ControlNameTextBox.BorderBrush;
        }

        public ToolboxControl(ToolboxControl t)
        {
            InitializeComponent();
            ID = IDGenerator.NextID();
            ImageSource = t.ImageSource;
            borderBrush = ControlNameTextBox.BorderBrush;
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);
            if (Mouse.LeftButton == MouseButtonState.Pressed && mouseState == MainWindow.MouseStateBeforeEnter)
            {
                // Package the data.
                DataObject data = new DataObject();
                data.SetData("Object", this);

                // Inititate the drag-and-drop operation.
                DragDrop.DoDragDrop(this, data, DragDropEffects.Copy);
            }
        }

        protected override void OnMouseEnter(MouseEventArgs e)
        {
            base.OnMouseEnter(e);
            if (Mouse.LeftButton == MouseButtonState.Pressed)
                mouseState = "";
            else
                mouseState = MainWindow.MouseStateBeforeEnter;
        }

        protected override void OnGiveFeedback(GiveFeedbackEventArgs e)
        {
            base.OnGiveFeedback(e);
            // These Effects values are set in the drop target's
            // DragOver event handler.
            if (e.Effects.HasFlag(DragDropEffects.Copy))
                Mouse.SetCursor(Cursors.Cross);
            else
                Mouse.SetCursor(Cursors.No);
            e.Handled = true;
        }

        protected override void OnDragOver(DragEventArgs e)
        {
            base.OnDragOver(e);
            e.Effects = DragDropEffects.Copy;
        }

        protected override void OnDrop(DragEventArgs e)
        {
            base.OnDrop(e);

            if (e.Data.GetDataPresent(MainWindow.LinkControlName))
            {
                LinkControl link = (LinkControl)e.Data.GetData(MainWindow.LinkControlName);

                string side = (string)e.Data.GetData("string");
                Binding b;
                if (side == "start")
                {
                    var endBindingItem = link.end.GetBindingExpression(Canvas.LeftProperty);
                    // If the other side is not connected to the same control.
                    if (endBindingItem == null || (endBindingItem != null && endBindingItem.DataItem != this))
                    {
                        b = new Binding()
                        {
                            Source = this,
                            Path = new PropertyPath(Canvas.LeftProperty),
                            Mode = BindingMode.TwoWay,
                            Converter = new OffsetConverter(),
                            ConverterParameter = ActualWidth / 2
                        };
                        link.start.SetBinding(Canvas.LeftProperty, b);

                        b = new Binding()
                        {
                            Source = this,
                            Path = new PropertyPath(Canvas.TopProperty),
                            Mode = BindingMode.TwoWay,
                            Converter = new OffsetConverter(),
                            ConverterParameter = ActualHeight / 2
                        };
                        link.start.SetBinding(Canvas.TopProperty, b);

                        link.IsStartConnected = true;
                        link.IDStart = ID;
                        if (link.IsEndConnected)
                        {
                            LinkControl.SetMyOffset(link);
                        }
                    }
                    // If we are trying to connect to a control that is already connected to the other side,
                    // move this side somewhere away from the control.
                    else
                    {
                        Canvas.SetLeft(link.start, Canvas.GetLeft(link.start));
                        Canvas.SetTop(link.start, Canvas.GetTop(link.start));
                    }
                }
                // If side == "end".
                else
                {
                    var startBindingItem = link.start.GetBindingExpression(Canvas.LeftProperty);
                    // If the other side is not connected to the same control.
                    if (startBindingItem == null || (startBindingItem != null && startBindingItem.DataItem != this))
                    {
                        b = new Binding()
                        {
                            Source = this,
                            Path = new PropertyPath(Canvas.LeftProperty),
                            Mode = BindingMode.TwoWay,
                            Converter = new OffsetConverter(),
                            ConverterParameter = ActualWidth / 2
                        };
                        link.end.SetBinding(Canvas.LeftProperty, b);

                        b = new Binding()
                        {
                            Source = this,
                            Path = new PropertyPath(Canvas.TopProperty),
                            Mode = BindingMode.TwoWay,
                            Converter = new OffsetConverter(),
                            ConverterParameter = ActualHeight / 2
                        };
                        link.end.SetBinding(Canvas.TopProperty, b);

                        link.IsEndConnected = true;
                        link.IDEnd = ID;
                        if (link.IsStartConnected)
                        {
                            LinkControl.SetMyOffset(link);
                        }
                    }
                    // If we are trying to connect to a control that is already connected to the other side,
                    // move this side somewhere away from the control.
                    else
                    {
                        Canvas.SetLeft(link.end, Canvas.GetLeft(link.end));
                        Canvas.SetTop(link.end, Canvas.GetTop(link.end));
                    }
                }
                ConfigurationAreaControl.IsChanged = true;
            }
        }

        protected bool AreAllNICsOccupied(ObservableCollection<NetworkInterfaceCard> interfaceCards)
        {
            foreach (var item in interfaceCards)
            {
                if (item.PluggedLink == null)
                    return false;
            }
            return true;
        }

        private void DeleteControl_Click(object sender, RoutedEventArgs e)
        {

            MainWindow window = (MainWindow)Application.Current.MainWindow;
            Canvas drawingArea = window.configurationArea.DrawingArea;

            // Get all link Controls.
            var elements = drawingArea.Children;
            DeleteToolboxControl(elements, this);
        }

        private void DeleteToolboxControl(UIElementCollection elements, ToolboxControl tc)
        {
            var counter = 0;
            var flows = mainWindow.configurationArea.Flows;
            // true == delete this and the Flow that it is part of.
            // false == stop here.
            if (IsPartOfFlow(flows))
            {
                // Ask user if he wants to proceed.
                var result = MessageBox.Show("This equipment is part of a flow/s.\nIf you delete it the flow/s will also be" +
                    " deleted. Do you want to proceed?", "", MessageBoxButton.YesNo, MessageBoxImage.Question);
                if (result == MessageBoxResult.No)
                    return;
            }

            // Remove the Flow object/s which contain this equipment.
            // All Flows that go through this equipment.
            var myComms = flows.FindAll(c => c.PathList.Find(t => t.Item1 == ID) != null);

            foreach (var comm in myComms)
            {
                var links = mainWindow.configurationArea.GetLinksPartOfFlow(comm);
                foreach (var link in links)
                {
                    link.UsedBandwidth -= comm.Throughput;
                }
            }
            flows.RemoveAll(c => myComms.Contains(c));

            // Remove the links that are bound to this equipment.
            foreach (var element in elements)
            {
                if (element.GetType().Name == MainWindow.LinkControlName)
                {
                    var ec = element as LinkControl;
                    var startBindingExpression = ec.start.GetBindingExpression(Canvas.LeftProperty);

                    // Find the ones that are bound to this Toolbox Control.
                    if (startBindingExpression != null && startBindingExpression.DataItem == tc)
                    {
                        // Modify and unbind their position.
                        var left = Canvas.GetLeft(ec.start);
                        var top = Canvas.GetTop(ec.start);

                        ec.start.SetBinding(Canvas.LeftProperty, "");
                        ec.start.SetBinding(Canvas.TopProperty, "");

                        ec.IsStartConnected = false;
                        ec.IDStart = -1;
                        ec.PortIndexStart = -1;

                        Canvas.SetLeft(ec.start, left + 15 * counter);
                        Canvas.SetTop(ec.start, top + 60);

                        ++counter;
                    }

                    var endBindingExpression = ec.end.GetBindingExpression(Canvas.LeftProperty);

                    if (endBindingExpression != null && endBindingExpression.DataItem == tc)
                    {
                        // Modify and unbind their position.
                        var left = Canvas.GetLeft(ec.end);
                        var top = Canvas.GetTop(ec.end);

                        ec.end.SetBinding(Canvas.LeftProperty, "");
                        ec.end.SetBinding(Canvas.TopProperty, "");

                        ec.IsEndConnected = false;
                        ec.IDEnd = -1;
                        ec.PortIndexEnd = -1;

                        Canvas.SetLeft(ec.end, left + 15 * counter);
                        Canvas.SetTop(ec.end, top + 60);

                        ++counter;
                    }
                }
            }
            // Remove the control.
            IDGenerator.FreeID(tc.ID);
            elements.Remove(tc);
        }

        private bool IsPartOfFlow(List<Flow> flows)
        {
            return flows.Find(c => c.PathList.Find(t => t.Item1 == ID) != null) != null ? true : false;
        }

        private async void ToolboxControl_MouseUp(object sender, MouseButtonEventArgs e)
        {
            if (e.ChangedButton == MouseButton.Left && !ControlNameTextBlock.IsMouseOver)
            {
                var b = sender as Border;
                Keyboard.Focus(b);
                e.Handled = true;

                ConfigurationAreaControl configurationArea = ((MainWindow)Application.Current.MainWindow).configurationArea;
                if (Keyboard.IsKeyDown(Key.LeftCtrl) && this is PC)
                {
                    if (configurationArea.Source == null)
                        configurationArea.Source = this;
                    else if (this != configurationArea.Source)
                    {
                        configurationArea.Destination = this;
                        await configurationArea.SetFlow();
                        configurationArea.ResetElementStyle();
                    }
                }
            }
        }

        protected override void OnKeyUp(KeyEventArgs e)
        {
            base.OnKeyUp(e);

            MainWindow window = (MainWindow)Application.Current.MainWindow;
            Canvas drawingArea = window.configurationArea.DrawingArea;

            if (e.Key == Key.Delete && Keyboard.FocusedElement == sp)
            {
                DeleteToolboxControl(drawingArea.Children, this);
            }
        }

        private void TextBlock_MouseDown(object sender, MouseButtonEventArgs e)
        {
            if (e.ClickCount == 2)
            {
                ControlNameTextBox.Visibility = Visibility.Visible;
                ControlNameTextBlock.Visibility = Visibility.Collapsed;
                ControlNameTextBox.Focus();
                ControlNameTextBox.CaretIndex = 10;
                e.Handled = true;
                originalName = ControlName;
            }
        }

        private void ControlNameTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            TextBox t = sender as TextBox;
            if (t.Text.Length > 0)
                t.BorderBrush = borderBrush;
            else
                t.BorderBrush = Brushes.Red;
        }

        private void ControlNameTextBox_LostFocus(object sender, RoutedEventArgs e)
        {
            ControlNameTextBlock.Visibility = Visibility.Visible;
            ControlNameTextBox.Visibility = Visibility.Collapsed;
            if (ControlNameTextBox.Text == "")
                ControlNameTextBox.Text = originalName;
        }
    }
}
